@@ -17,6 +17,7 @@ sub startup {
 	my $dir = $ENV{SREVIEW_WDIR};
 
 	$self->config(hypnotoad => { pid_file => '/var/run/sreview/sreview-web.pid' });
+	$self->plugin('OAuth2::Server' => 
 
 	my $config = SReview::Config::Common::setup;
 
@@ -44,7 +45,7 @@ sub startup {
 		if(defined($vpr) && length($vpr) > 0) {
 			$media = "media-src $vpr;";
 		}
-		$c->res->headers->content_security_policy("default-src 'none'; script-src 'self' 'unsafe-inline'; font-src 'self'; style-src 'self'; img-src 'self'; frame-ancestors 'none'; $media");
+		$c->res->headers->content_security_policy("default-src 'none'; connect-src 'self'; script-src 'self' 'unsafe-inline'; font-src 'self'; style-src 'self'; img-src 'self'; frame-ancestors 'none'; $media");
 	});
 
 	$self->helper(dbh => sub {
@@ -154,6 +155,10 @@ sub startup {
 		}
 	});
 
+	$r->get('/r/:nonce')->to(controller => 'review', action => 'view', layout => 'default');
+	$r->post('/r/:nonce/update')->to(controller => 'review', layout => 'default', action => 'update');
+        $r->get('/r/:nonce/data')->to(controller => 'review', action => 'data');
+
 	$r->get('/review/:nonce' => sub {
 		my $c = shift;
 		my $stt = $c->dbh->prepare("SELECT state, name, id, extract(epoch from prelen) as prelen, extract(epoch from postlen) as postlen, extract(epoch from (endtime - starttime)) as length, speakers, starttime, endtime, slug, room, comments FROM talk_list WHERE nonce=?");
@@ -299,6 +304,7 @@ sub startup {
 		$expls{'preview'} = 'Talk ready for review, waiting for reviewer';
 		$expls{'transcoding'} = 'High-quality transcodes running';
 		$expls{'uploading'} = 'Publishing results';
+		$expls{'announcing'} = 'Announcing completion of publication';
 		$expls{'done'} = 'Videos published, all done';
 		$expls{'broken'} = 'Review found problems, administrator required';
 		$expls{'ignored'} = 'Talk will not be/was not recorded, ignored for review';
@@ -315,7 +321,7 @@ sub startup {
 			foreach my $row(@$rows) {
 				my $nonce = shift @$row;
 				my $name = shift @$row;
-				push @$newrows, [ "<a href='/review/$nonce'>$name</a>", @$row ];
+				push @$newrows, [ "<a href='/r/$nonce'>$name</a>", @$row ];
 			}
 			$rows = $newrows;
 		}
@@ -331,10 +337,30 @@ sub startup {
 		$c->render;
 	} => 'table');
 
-	$r->post('/talk_update')->to(controller => 'talk', action => 'update',
-			layout => 'default', name => 'talk_update',
-			notfound_message => "Unauthorized.",
-			notfound_code => 403);
+	$r->post('/talk_update' => sub {
+		my $c = shift;
+		my $nonce = $c->param("nonce");
+		if(!defined($nonce)) {
+			$c->stash(message=>"Unauthorized.");
+			$c->res->code(403);
+			$c->render('error');
+			return undef;
+		}
+		my $sth = $c->dbh->prepare("SELECT id FROM talks WHERE nonce = ? AND state IN ('preview', 'broken')");
+		$sth->execute($nonce);
+		my $row = $sth->fetchrow_arrayref;
+		if(scalar($row) == 0) {
+			$c->stash(message=>"Change not allowed. If this talk exists, it was probably reviewed by someone else while you were doing so too. Please try again later, or check the overview page.");
+			$c->res->code(403);
+			$c->render('error');
+			return undef;
+		}
+		$c->stash(layout => 'default');
+		$c->stash(template => 'talk');
+		$c->flash(completion_message => 'Your change has been accepted. Thanks for your help!');
+		$c->talk_update($row->[0]);
+		$c->redirect_to("/review/$nonce");
+	} => 'talk_update');
 
 	my $vol = $r->under('/volunteer' => sub {
 		my $c = shift;
@@ -433,14 +459,23 @@ sub startup {
 		$c->render(template => 'talk');
 	} => 'admin_talk');
 
-	$admin->post('/talk_update')->to(controller => 'talk', action => 'update',
-			name => 'talk_update_admin',
-			notfound_message => "Required parameter talk missing",
-			notfound_code => 404);
+	$admin->post('/talk_update' => sub {
+		my $c = shift;
+		my $talk = $c->param("talk");
+		if(!defined($talk)) {
+			$c->stash(message => "Required parameter talk missing.");
+			$c->render("error");
+			return undef;
+		}
+		$c->stash(template => 'talk');
+		$c->flash(completion_message => 'Your change has been accepted. Thanks for your help!');
+		$c->talk_update($talk);
+		$c->redirect_to("/admin/talk?talk=$talk");
+	} => 'talk_update_admin');
 
 	$admin->get('/brokens' => sub {
 		my $c = shift;
-		my $st = $c->dbh->prepare("SELECT talks.id, title, speakeremail(talks.id), tracks.email, comments, state FROM talks JOIN tracks ON talks.track = tracks.id WHERE state>='broken' ORDER BY state,id");
+		my $st = $c->dbh->prepare("SELECT talks.id, title, speakeremail(talks.id), tracks.email, comments, state, nonce FROM talks JOIN tracks ON talks.track = tracks.id WHERE state>='broken' ORDER BY state,id");
 		my $tst = $c->dbh->prepare("SELECT rooms.altname, count(talks.id) FROM talks JOIN rooms ON talks.room = rooms.id WHERE talks.state='broken' GROUP BY rooms.altname");
 		my $rows = [];
 		$st->execute;
@@ -450,7 +485,8 @@ sub startup {
 		$c->stash(tottitrow => [ 'Room', 'Count' ]);
 		my $pgrows = $st->fetchall_arrayref;
 		foreach my $row(@{$pgrows}) {
-			push @$row, "<a href='/admin/talk?talk=" . $row->[0] . "'>review</a>";
+                        my $nonce = pop @$row;
+			push @$row, "<a href='/r/$nonce'>review</a>";
 			push @$rows, $row;
 		}
 		$c->stash(rows => $rows);
